(* Module ColFormat *)
(* ----------------
 * Columns pretty-printer
 *)

module String = More.String

(*Type colonne : nbLignes*nbColonnes*liste de lignes*)

type column = int * int * (line list)
 and line =
   | Center of string
   | Left of string
   | Cols of int * column list
   | Empty

let new_column = (0, 0, [])

let append_left (w, h, lines) txt =
  (max w (String.utf8_length txt), h + 1, (Left txt)::lines)

let append_center (w, h, lines) txt =
  (max w (String.utf8_length txt), h + 1, (Center txt)::lines)

let rec fill_col to_size ((w, h, lines) as col) =
  if h < to_size
    then fill_col to_size (w, h+1, Empty::lines)
    else col

let append_cols (w, h, lines) cols =
  let (cw, ch) = List.fold_left
                  (fun (rw,rh) (cw, ch, _) -> (rw + cw, max rh ch))
                  (0, 0) cols
  in let aux (w, h, lines) = fill_col ch (w, h, List.rev lines)
  in (  max w (cw + List.length cols - 1),
        h + ch,
        (Cols (cw, List.map aux cols)::lines)
     )

let join cols =
  let (cw, ch) = List.fold_left
                  (fun (rw,rh) (cw, ch, _) -> (rw + cw, max rh ch))
                  (0, 0) cols
  in let aux (w, h, lines) = fill_col ch (w, h, List.rev lines)
  in (  cw + List.length cols - 1,
        ch,
        [Cols (cw, List.map aux cols)]
     )

let spaces n = String.make n ' '

let rec _print_cols ff w lst =
  match lst with
    | [] -> ()
    | l ->
        let ret = _print_line ff w l in
        Format.pp_print_cut ff () ;
        _print_cols ff w ret

and _print_line ff w = function
  | [] -> []

  | (Center txt)::q ->
      let len = String.utf8_length txt in
      Format.pp_print_string ff (spaces ((w - len) / 2)) ;
      Format.pp_print_string ff txt ;
      Format.pp_print_string ff (spaces ((w + 1 - len) / 2)) ;
      q

  | (Left txt)::q ->
      let len = String.utf8_length txt in
      Format.pp_print_string ff txt ;
      Format.pp_print_string ff (spaces (w - len)) ;
      q

  | (Cols (_, []))::q ->
      _print_line ff w q

  | (Cols (_, (_,0,_)::_))::q ->
      _print_line ff w q

  | (Cols (cw,l))::q ->
      let len = ref (List.length l - 1) in
      let spc = ref (w - cw) in
      let trans (w,h, lst) =
        let ret = (w, h-1, _print_line ff w lst)
        and nbspc = if !len > 0 then !spc / !len else 0 in
        Format.pp_print_string ff (spaces nbspc) ;
        spc := !spc - nbspc ;
        len := !len - 1 ;
        ret
      in (Cols (cw, List.map trans l)::q)

    | Empty::q -> Format.pp_print_string ff (spaces w) ; q

let printer ff (w, h, lines) =
  Format.pp_open_vbox ff 0 ;
  _print_cols ff w (List.rev lines) ;
  Format.pp_close_box ff ()
