(** Indexed structures using functions. *)

type 'a t = int * (int -> 'a)

let empty = (0, fun _ -> raise Not_found)

let single v =
  (1, function 0 -> v | _ -> raise Not_found)

let concat (n_a, f_a) (n_b, f_b) =
  (n_a + n_b, fun x -> if x < n_a then f_a x else f_b (x - n_a))

let is_empty (n, _) = n = 0

let length = fst

let equal (na,fa) (nb,fb) =
  let rec aux i =
    if i < 0 then true else
    if fa i = fb i then aux (i-1)
    else false
  in
  if na = nb then aux (na - 1) else false

let nth = snd

let find_first test beg (_,f) =
  let rec aux pos =
    if test (f pos)
      then pos
      else aux (pos + 1)
  in try (
    aux beg
  ) with
    | Not_found -> -1

let replace f_replace (n, f) pos =
  (n, fun x -> if x = pos then f_replace (f x) else f x)

let remove (n, f) pos =
  if n = 0
    then empty
    else (n - 1, fun x -> if x < pos then f x else f (x + 1))

let map f_map (n, f) =
  (n, fun x -> f_map (f x))

let fold_left f_arg acc (n, f) =
  let rec aux acc pos =
    try (
      aux (f_arg acc (f pos)) (pos + 1)
    ) with
      | Not_found -> acc
  in aux acc 0

let from_list l =
  ( List.length l,
    (fun i -> try (List.nth l i) with (Failure "nth") -> raise Not_found) )

let to_list (n,f) =
  let rec aux acc i =
    if i < 0 then acc else
    aux ((f i)::acc) (i - 1)
  in aux [] (n - 1)
