(** Polymorphic 1-2-trees. *)

type ('a, 'b, 'c) tree =
  | Leaf   of 'a
  | Unary  of 'b * ('a, 'b, 'c) tree
  | Binary of 'c * ('a, 'b, 'c) tree * ('a, 'b, 'c) tree

let fold_down f_leaf f_unary f_binary tree =
  (* Really simple implementation. Should probably be tail recursive. *)
  let rec aux = function
    | Leaf    v        -> f_leaf   v
    | Unary  (v, a)    -> f_unary  (aux a) v
    | Binary (v, a, b) -> f_binary (aux a) (aux b) v
  in aux tree

type ('s, 'u, 'b) prooftree = ('s, 'u * 's, 'b * 's) tree

let seq_from_prooftree = function
    | Leaf v -> v
    | Unary ((_, v), _) -> v
    | Binary ((_,v),_,_) -> v

let pseudo_map_down f_unary f_binary tree =
  (* Really simple implementation. Should probably be tail recursive. *)
  let rec aux = function
    | Leaf v -> Leaf v
    | Unary (w, a) ->
        let na = aux a in
        Unary (f_unary (seq_from_prooftree na) w, na)
    | Binary (w, a, b) ->
        let na = aux a
        and nb = aux b in
        Binary (f_binary (seq_from_prooftree na) (seq_from_prooftree nb) w,
                na, nb)
  in aux tree

let map_down f_leaf f_unary f_binary tree =
  (* Really simple implementation. Should probably be tail recursive. *)
  let rec aux = function
    | Leaf    v        -> Leaf   (f_leaf   v)
    | Unary  (v, a)    -> Unary  (f_unary  v, aux a)
    | Binary (v, a, b) -> Binary (f_binary v, aux a, aux b)
  in aux tree

let replace old_branch new_branch =
  let rec aux = function
    | z when z = old_branch -> new_branch
    | Leaf v           -> raise Not_found
    | Unary  (v, a)    -> Unary (v, aux a)
    | Binary (v, a, b) ->
      ( try               Binary (v, aux a, b)
        with Not_found -> Binary (v, a, aux b)
      )
  in aux 
